#include <set>
#include <fstream>
#include <sstream>
#include <time.h>

#include "common.h"
#include "tasedit.h"
#include "taseditlib/taseditproj.h"
#include "fceu.h"
#include "debugger.h"
#include "replay.h"
#include "movie.h"
#include "utils/xstring.h"
#include "Win32InputBox.h"
#include "window.h"
#include "keyboard.h"
#include "joystick.h"
#include "help.h"
#include "main.h"	//For the GetRomName() function

using namespace std;

//to change header font
//http://forums.devx.com/archive/index.php/t-37234.html

bool old_project_changed;
int old_multitrack_recording_joypad, multitrack_recording_joypad;
bool old_movie_readonly;
int lastCursor;
bool old_emu_paused, emu_paused;
int old_pauseframe;
bool old_show_pauseframe, show_pauseframe;
bool old_rewind_button_state, rewind_button_state;
bool old_forward_button_state, forward_button_state;
int button_hold_time;
int seeking_start_frame = 0;
bool TASEdit_focus = false;
int listItems;	// number of items in list
// saved FCEU config
int saved_eoptions;
int saved_EnableAutosave;
extern int EnableAutosave;
bool saved_compressSavestates;
extern bool compressSavestates;

// vars saved in cfg file
int TasEdit_wndx, TasEdit_wndy;
bool TASEdit_follow_playback = true;
bool TASEdit_show_lag_frames = true;
bool TASEdit_show_markers = true;
bool TASEdit_bind_markers = true;
bool TASEdit_restore_position = false;
int TASEdit_greenzone_capacity = GREENZONE_DEFAULT_CAPACITY;
extern bool muteTurbo;
bool TASEdit_show_dot = true;

string tasedithelp = "{16CDE0C4-02B0-4A60-A88D-076319909A4D}"; //Name of TASEdit Help page
char buttonNames[NUM_JOYPAD_BUTTONS][2] = {"A", "B", "S", "T", "U", "D", "L", "R"};
char windowCaptions[6][30] = {	"TAS Editor",
								"TAS Editor (Recording All)",
								"TAS Editor (Recording 1P)",
								"TAS Editor (Recording 2P)",
								"TAS Editor (Recording 3P)",
								"TAS Editor (Recording 4P)"};
HWND hwndTasEdit = 0;
static HMENU hmenu, hrmenu;
static HWND hwndList, hwndHeader, hwndProgressbar, hwndRewind, hwndForward;
static HWND hwndRB_RecOff, hwndRB_RecAll, hwndRB_Rec1P, hwndRB_Rec2P, hwndRB_Rec3P, hwndRB_Rec4P;
static WNDPROC hwndHeader_oldWndproc, hwndList_oldWndProc;

typedef std::set<int> TSelectionFrames;
static TSelectionFrames selectionFrames;

//hacky.. we need to think about how to convey information from the driver to the movie code.
//add a new fceud_ function?? blehhh maybe
extern EMOVIEMODE movieMode;

TASEDIT_PROJECT project;

static void GetDispInfo(NMLVDISPINFO* nmlvDispInfo)
{
	LVITEM& item = nmlvDispInfo->item;
	if(item.mask & LVIF_TEXT)
	{
		switch(item.iSubItem)
		{
			case COLUMN_ICONS:
			if(item.iImage == I_IMAGECALLBACK && item.iItem == currFrameCounter)
				item.iImage = 0;
			else
				item.iImage = -1;
			break;
			case COLUMN_FRAMENUM:
			case COLUMN_FRAMENUM2:
			{
				U32ToDecStr(item.pszText,item.iItem,DIGITS_IN_FRAMENUM);
				break;
			}
			case COLUMN_JOYPAD1_A: case COLUMN_JOYPAD1_B: case COLUMN_JOYPAD1_S: case COLUMN_JOYPAD1_T:
			case COLUMN_JOYPAD1_U: case COLUMN_JOYPAD1_D: case COLUMN_JOYPAD1_L: case COLUMN_JOYPAD1_R:
			case COLUMN_JOYPAD2_A: case COLUMN_JOYPAD2_B: case COLUMN_JOYPAD2_S: case COLUMN_JOYPAD2_T:
			case COLUMN_JOYPAD2_U: case COLUMN_JOYPAD2_D: case COLUMN_JOYPAD2_L: case COLUMN_JOYPAD2_R:
			case COLUMN_JOYPAD3_A: case COLUMN_JOYPAD3_B: case COLUMN_JOYPAD3_S: case COLUMN_JOYPAD3_T:
			case COLUMN_JOYPAD3_U: case COLUMN_JOYPAD3_D: case COLUMN_JOYPAD3_L: case COLUMN_JOYPAD3_R:
			case COLUMN_JOYPAD4_A: case COLUMN_JOYPAD4_B: case COLUMN_JOYPAD4_S: case COLUMN_JOYPAD4_T:
			case COLUMN_JOYPAD4_U: case COLUMN_JOYPAD4_D: case COLUMN_JOYPAD4_L: case COLUMN_JOYPAD4_R:
			{
				int joy = (item.iSubItem - COLUMN_JOYPAD1_A) / NUM_JOYPAD_BUTTONS;
				int bit = (item.iSubItem - COLUMN_JOYPAD1_A) % NUM_JOYPAD_BUTTONS;
				uint8 data = currMovieData.records[item.iItem].joysticks[joy];
				if(data & (1<<bit))
				{
					item.pszText[0] = MovieRecord::mnemonics[bit];
					item.pszText[1] = 0;
				} else
				{
					if (TASEdit_show_dot)
					{
						item.pszText[0] = 46;	// "."
						item.pszText[1] = 0;
					} else item.pszText[0] = 0;
				}
			}
			break;
		}
	}
}

#define CDDS_SUBITEMPREPAINT       (CDDS_SUBITEM | CDDS_ITEMPREPAINT)
#define CDDS_SUBITEMPOSTPAINT      (CDDS_SUBITEM | CDDS_ITEMPOSTPAINT)
#define CDDS_SUBITEMPREERASE       (CDDS_SUBITEM | CDDS_ITEMPREERASE)
#define CDDS_SUBITEMPOSTERASE      (CDDS_SUBITEM | CDDS_ITEMPOSTERASE)

static LONG CustomDraw(NMLVCUSTOMDRAW* msg)
{
	//LPNMCUSTOMDRAW

	int cell_x, cell_y;
	switch(msg->nmcd.dwDrawStage)
	{
	case CDDS_PREPAINT:
		return CDRF_NOTIFYITEMDRAW;
	case CDDS_ITEMPREPAINT:
		return CDRF_NOTIFYSUBITEMDRAW;
	case CDDS_SUBITEMPREPAINT:
		SelectObject(msg->nmcd.hdc,debugSystem->hFixedFont);
		cell_x = msg->iSubItem;
		cell_y = msg->nmcd.dwItemSpec;

		//msg->clrText = 0xFF0000;

		if(cell_x > COLUMN_ICONS)
		{
			if(cell_x == COLUMN_FRAMENUM || cell_x == COLUMN_FRAMENUM2)
			{
				// frame number
				if (cell_y == currFrameCounter || (cell_y == pauseframe-1 && show_pauseframe))
				{
					// current frame
					if(TASEdit_show_markers && currMovieData.frames_flags[cell_y] & MARKER_FLAG_BIT)
						// this frame is also marked
						msg->clrTextBk = CUR_MARKED_FRAMENUM_COLOR;
					else
						msg->clrTextBk = CUR_FRAMENUM_COLOR;
				} else if(TASEdit_show_markers && currMovieData.frames_flags[cell_y] & MARKER_FLAG_BIT)
				{
					// marked frame
					msg->clrTextBk = MARKED_FRAMENUM_COLOR;
				} else if(cell_y < currMovieData.greenZoneCount && !currMovieData.savestates[cell_y].empty())
				{
					if (TASEdit_show_lag_frames && (currMovieData.frames_flags[cell_y] & LAG_FLAG_BIT))
					{
						// lag frame
						msg->clrTextBk = LAG_FRAMENUM_COLOR;
					} else
					{
						// green zone frame
						msg->clrTextBk = GREENZONE_FRAMENUM_COLOR;
					}
				} else msg->clrTextBk = NORMAL_FRAMENUM_COLOR;
			} else if((cell_x - COLUMN_JOYPAD1_A) / NUM_JOYPAD_BUTTONS == 0 || (cell_x - COLUMN_JOYPAD1_A) / NUM_JOYPAD_BUTTONS == 2)
			{
				// pad 1 or 3
				if (cell_y == currFrameCounter || (cell_y == pauseframe-1 && show_pauseframe))
				{
					// current frame
					msg->clrTextBk = CUR_INPUT_COLOR1;
				} else if(cell_y < currMovieData.greenZoneCount && !currMovieData.savestates[cell_y].empty())
				{
					if (TASEdit_show_lag_frames && (currMovieData.frames_flags[cell_y] & LAG_FLAG_BIT))
					{
						// lag frame
						msg->clrTextBk = LAG_INPUT_COLOR1;
					} else
					{
						// green zone frame
						msg->clrTextBk = GREENZONE_INPUT_COLOR1;
					}
				} else msg->clrTextBk = NORMAL_INPUT_COLOR1;
			} else if((cell_x - COLUMN_JOYPAD1_A) / NUM_JOYPAD_BUTTONS == 1 || (cell_x - COLUMN_JOYPAD1_A) / NUM_JOYPAD_BUTTONS == 3)
			{
				// pad 2 or 4
				if (cell_y == currFrameCounter || (cell_y == pauseframe-1 && show_pauseframe))
				{
					// current frame
					msg->clrTextBk = CUR_INPUT_COLOR2;
				} else if(cell_y < currMovieData.greenZoneCount && !currMovieData.savestates[cell_y].empty())
				{
					if (TASEdit_show_lag_frames && (currMovieData.frames_flags[cell_y] & LAG_FLAG_BIT))
					{
						// lag frame
						msg->clrTextBk = LAG_INPUT_COLOR2;
					} else
					{
						// green zone frame
						msg->clrTextBk = GREENZONE_INPUT_COLOR2;
					}
				} else msg->clrTextBk = NORMAL_INPUT_COLOR2;
			}
		}
		return CDRF_DODEFAULT;
	default:
		return CDRF_DODEFAULT;
	}
}

// called from the rest of the emulator when things happen and the tasedit should change to reflect it
void UpdateTasEdit()
{
	if(!hwndTasEdit) return;
	
	// pause when seeking hit pauseframe
	if(!FCEUI_EmulationPaused())
		if(pauseframe && pauseframe <= currFrameCounter + 1)
			SeekingStop();

	// update seeking progressbar
	old_emu_paused = emu_paused;
	emu_paused = (bool)FCEUI_EmulationPaused();
	if (pauseframe)
	{
		SendMessage(hwndProgressbar, PBM_SETPOS, PROGRESSBAR_WIDTH * (currFrameCounter-seeking_start_frame) / (pauseframe-seeking_start_frame), 0);
	} else if (old_emu_paused != emu_paused)
	{
		// emulator got paused/unpaused externally
		if (old_emu_paused && !emu_paused)
			// externally unpaused - progressbar should be empty
			SendMessage(hwndProgressbar, PBM_SETPOS, 0, 0);
		else
			// externally paused - progressbar should be full
			SendMessage(hwndProgressbar, PBM_SETPOS, PROGRESSBAR_WIDTH, 0);
	}
	// update flashing pauseframe
	if (old_pauseframe != pauseframe && old_pauseframe) RedrawRow(old_pauseframe-1);
	old_pauseframe = pauseframe;
	old_show_pauseframe = show_pauseframe;
	if (pauseframe)
		show_pauseframe = (int)(clock() / PAUSEFRAME_BLINKING_PERIOD) & 1;
	else
		show_pauseframe = false;
	if (old_show_pauseframe != show_pauseframe) RedrawRow(pauseframe-1);

	UpdateList();
	//UpdateRecordingRadioButtons();

	//update the cursor
	if(currFrameCounter != lastCursor)
	{
		FollowPlayback();
		//update the old and new rows
		RedrawRow(lastCursor);
		RedrawRow(currFrameCounter);
		UpdateWindow(hwndList);
		lastCursor = currFrameCounter;
	}
	
	// update < and > buttons
	if(emu_paused)
	{
		old_rewind_button_state = rewind_button_state;
		rewind_button_state = (bool)(Button_GetState(hwndRewind) & BST_PUSHED);
		if (rewind_button_state)
		{
			if (!old_rewind_button_state)
			{
				button_hold_time = clock();
				Tasedit_RewindFrame();
			} else if (button_hold_time + HOLD_REPEAT_DELAY < clock())
			{
				Tasedit_RewindFrame();
			}
		}
		old_forward_button_state = forward_button_state;
		forward_button_state = (bool)(Button_GetState(hwndForward) & BST_PUSHED);
		if (forward_button_state)
		{
			if (!old_forward_button_state)
			{
				button_hold_time = clock();
				Tasedit_ForwardFrame();
			} else if (button_hold_time + HOLD_REPEAT_DELAY < clock())
			{
				Tasedit_ForwardFrame();
			}
		}
	}

	// update window caption
	if (old_movie_readonly != movie_readonly || old_multitrack_recording_joypad != multitrack_recording_joypad || old_project_changed != project.changed)
		RedrawWindowCaption();
	old_project_changed = project.changed;

	// update recording radio buttons if user used hotkey to switch R/W
	if (old_movie_readonly != movie_readonly || old_multitrack_recording_joypad != multitrack_recording_joypad)
	{
		UncheckRecordingRadioButtons();
		RecheckRecordingRadioButtons();
	}

}

void UpdateList()
{
	//update the number of items in the list
	int currLVItemCount = ListView_GetItemCount(hwndList);
	if(currMovieData.getNumRecords() != currLVItemCount)
	{
		ListView_SetItemCountEx(hwndList,currMovieData.getNumRecords(),LVSICF_NOSCROLL | LVSICF_NOINVALIDATEALL);
	}
}

void UpdateProgressbar(int frame)
{
	SendMessage(hwndProgressbar, PBM_SETPOS, PROGRESSBAR_WIDTH * frame / currMovieData.greenZoneCount, 0);
}

void RedrawWindowCaption()
{
	char windowCaption[300];	// 260 + 30 + 1 + ...
	if (movie_readonly)
		strcpy(windowCaption, windowCaptions[0]);
	else
		strcpy(windowCaption, windowCaptions[multitrack_recording_joypad + 1]);
	// add project name
	std::string projectname = project.GetProjectName();
	if (!projectname.empty())
	{
		strcat(windowCaption, " - ");
		strcat(windowCaption, projectname.c_str());
	}
	// and * if project has unsaved changes
	if (project.changed)
		strcat(windowCaption, "*");
	SetWindowText(hwndTasEdit, windowCaption);
}
void RedrawTasedit()
{
	InvalidateRect(hwndTasEdit,0,FALSE);
}
void RedrawList()
{
	InvalidateRect(hwndList,0,FALSE);
}
void RedrawRow(int index)
{
	ListView_RedrawItems(hwndList,index,index);
}

void Tasedit_RewindFrame()
{
	if (currFrameCounter > 0) JumpToFrame(currFrameCounter-1);
	FollowPlayback();
}
void Tasedit_ForwardFrame()
{
	JumpToFrame(currFrameCounter+1);
	FollowPlayback();
}

void Tasedit_ToggleEmulationPause()
{
	if (FCEUI_EmulationPaused())
		UnpauseEmulation();
	else
		PauseEmulation();
}
void PauseEmulation()
{
	FCEUI_SetEmulationPaused(1);
	// make some additional stuff
}
void UnpauseEmulation()
{
	FCEUI_SetEmulationPaused(0);
	// make some additional stuff
}

void SeekingStart(int finish_frame)
{
	seeking_start_frame = currFrameCounter;
	pauseframe = finish_frame;
	turbo = (seeking_start_frame + FRAMES_TOO_FAR < finish_frame);
	UnpauseEmulation();
}
void SeekingStop()
{
	pauseframe = 0;
	turbo = false;
	PauseEmulation();
	SendMessage(hwndProgressbar, PBM_SETPOS, PROGRESSBAR_WIDTH, 0);
}

enum ECONTEXTMENU
{
	CONTEXTMENU_STRAY = 0,
	CONTEXTMENU_SELECTED = 1,
};

void ShowMenu(ECONTEXTMENU which, POINT& pt)
{
	HMENU sub = GetSubMenu(hrmenu,(int)which);
	TrackPopupMenu(sub,0,pt.x,pt.y,TPM_RIGHTBUTTON,hwndTasEdit,0);
}

void StrayClickMenu(LPNMITEMACTIVATE info)
{
	POINT pt = info->ptAction;
	ClientToScreen(hwndList,&pt);
	ShowMenu(CONTEXTMENU_STRAY,pt);
}

void RightClickMenu(LPNMITEMACTIVATE info)
{
	POINT pt = info->ptAction;
	ClientToScreen(hwndList,&pt);
	ShowMenu(CONTEXTMENU_SELECTED,pt);
}

void RightClick(LPNMITEMACTIVATE info)
{
	int index = info->iItem;
	int column = info->iSubItem;

	//stray clicks give a context menu:
	if(index == -1)
	{
		StrayClickMenu(info);
		return;
	}

	//make sure that the click is in our currently selected set.
	//if it is not, then we don't know what to do yet
	if(selectionFrames.find(index) == selectionFrames.end())
	{
		return;
	}
	
	RightClickMenu(info);
}

void InputChanged()
{
	// keep input log
	// update hot input


	project.changed = true;
}
void MarkersChanged()
{
	project.changed = true;
}

void InvalidateGreenZone(int after)
{
	if (currMovieData.greenZoneCount > after+1)
	{
		currMovieData.greenZoneCount = after+1;
		currMovieData.rerecordCount++;
		// either set playback cursor to the end of greenzone or run seeking to restore playback position
		if (currFrameCounter >= currMovieData.greenZoneCount)
		{
			if (TASEdit_restore_position)
			{
				if (pauseframe)
					JumpToFrame(pauseframe-1);
				else
					JumpToFrame(currFrameCounter);
				turbo = true;
			} else
			{
				JumpToFrame(currMovieData.greenZoneCount-1);
			}
		}
	}
	// redraw list even if greenzone didn't change
	RedrawList();
}

bool JumpToFrame(int index)
{
	// Returns true if a jump to the frame is made, false if nothing's done. 
	if (index<0) return false; 

	if (index >= currMovieData.greenZoneCount)
	{
		/* Handle jumps outside greenzone. */
		if (JumpToFrame(currMovieData.greenZoneCount-1))
		{
			// seek from the end of greenzone
			SeekingStart(index+1);
			return true;
		}
		return false;
	}
	/* Handle jumps inside greenzone. */
	if (currMovieData.loadTasSavestate(index))
	{
		currFrameCounter = index;
		turbo = false;
		// if playback was seeking, pause emulation right here
		if (pauseframe) SeekingStop();
		return true;
	}
	//Search for an earlier frame with savestate
	int i = (index>0)? index-1 : 0;
	for (; i > 0; --i)
	{
		if (currMovieData.loadTasSavestate(i)) break;
	}
	if (!i)
		StartFromZero();
	else
		currFrameCounter = i;
	// continue from the frame
	SeekingStart(index+1);
	return true;
}

void StartFromZero()
{
	poweron(true);
	currFrameCounter = 0;
	currMovieData.TryDumpIncremental();
}

void ToggleJoypadBit(int column_index, int row_index, UINT KeyFlags)
{
	int joy = (column_index - COLUMN_JOYPAD1_A) / NUM_JOYPAD_BUTTONS;
	int bit = (column_index - COLUMN_JOYPAD1_A) % NUM_JOYPAD_BUTTONS;
	if (KeyFlags & (LVKF_SHIFT|LVKF_CONTROL))
	{
		//update multiple rows 
		for(TSelectionFrames::iterator it(selectionFrames.begin()); it != selectionFrames.end(); it++)
		{
			currMovieData.records[*it].toggleBit(joy,bit);
		}
		row_index = *selectionFrames.begin();
	} 
	else 
	{
		//update one row
		currMovieData.records[row_index].toggleBit(joy,bit);
	}
	InputChanged();
	InvalidateGreenZone(row_index);
}

void SingleClick(LPNMITEMACTIVATE info)
{
	int row_index = info->iItem;
	if(row_index == -1) return;
	int column_index = info->iSubItem;

	if(column_index == COLUMN_ICONS)
	{
		// click on the "icons" column - jump to the frame
		ClearSelection();
		JumpToFrame(row_index);
		RedrawList();
	} else if(column_index == COLUMN_FRAMENUM || column_index == COLUMN_FRAMENUM2)
	{
		// click on the "frame number" column - set marker if clicked with Alt
		if (info->uKeyFlags & LVKF_ALT)
		{
			// reverse MARKER_FLAG_BIT in pointed frame
			if (currMovieData.frames_flags[row_index] & MARKER_FLAG_BIT)
				currMovieData.frames_flags[row_index] &= ~MARKER_FLAG_BIT;
			else
				currMovieData.frames_flags[row_index] |= MARKER_FLAG_BIT;
			MarkersChanged();
			ListView_SetItemState(hwndList,row_index,0,LVIS_SELECTED);
			//RedrawList();
		}
	}
	else if(column_index >= COLUMN_JOYPAD1_A && column_index <= COLUMN_JOYPAD4_R)
	{
		ToggleJoypadBit(column_index, row_index, info->uKeyFlags);
	}
}

void DoubleClick(LPNMITEMACTIVATE info)
{
	int row_index = info->iItem;
	if(row_index == -1) return;
	int column_index = info->iSubItem;

	if(column_index == COLUMN_ICONS || column_index == COLUMN_FRAMENUM || column_index == COLUMN_FRAMENUM2)
	{
		// double click sends playback to the frame
		ClearSelection();
		JumpToFrame(row_index);
		RedrawList();
	} else if(column_index >= COLUMN_JOYPAD1_A && column_index <= COLUMN_JOYPAD4_R)
	{
		ToggleJoypadBit(column_index, row_index, info->uKeyFlags);
	}
}

void CloneFrames()
{
	int frames = selectionFrames.size();

	currMovieData.records.reserve(currMovieData.records.size()+frames);
	currMovieData.frames_flags.reserve(currMovieData.frames_flags.size()+frames);

	//insert frames before each selection, but consecutive selection lines are accounted as single region
	frames = 1;
	TSelectionFrames::reverse_iterator next_it;
	for(TSelectionFrames::reverse_iterator it(selectionFrames.rbegin()); it != selectionFrames.rend(); it++)
	{
		next_it = it;
		next_it++;
		if (next_it == selectionFrames.rend() || (int)*next_it < ((int)*it - 1))
		{
			// end of current region
			currMovieData.cloneRegion(*it,frames);
			frames = 1;
		} else frames++;
	}

	UpdateList();
	InputChanged();
	InvalidateGreenZone(*selectionFrames.begin());
}

void InsertFrames()
{
	int frames = selectionFrames.size();

	//this is going to be slow.

	//to keep this from being even slower than it would otherwise be, go ahead and reserve records
	currMovieData.records.reserve(currMovieData.records.size()+frames);
	currMovieData.frames_flags.reserve(currMovieData.frames_flags.size()+frames);

	//insert frames before each selection, but consecutive selection lines are accounted as single region
	frames = 1;
	TSelectionFrames::reverse_iterator next_it;
	for(TSelectionFrames::reverse_iterator it(selectionFrames.rbegin()); it != selectionFrames.rend(); it++)
	{
		next_it = it;
		next_it++;
		if (next_it == selectionFrames.rend() || (int)*next_it < ((int)*it - 1))
		{
			// end of current region
			currMovieData.insertEmpty(*it,frames);
			frames = 1;
		} else frames++;
	}

	UpdateList();
	InputChanged();
	InvalidateGreenZone(*selectionFrames.begin());
}

void DeleteFrames()
{
	//delete frames on each selection, going backwards
	for(TSelectionFrames::reverse_iterator it(selectionFrames.rbegin()); it != selectionFrames.rend(); it++)
	{
		currMovieData.records.erase(currMovieData.records.begin() + *it);
		if (TASEdit_bind_markers)
			currMovieData.frames_flags.erase(currMovieData.frames_flags.begin() + *it);
	}
	// check if user deleted all frames
	if (!currMovieData.records.size())
		StartFromZero();
	// reduce list
	InputChanged();
	UpdateList();
	int index = *selectionFrames.begin();
	int delete_index;
	// reduce selection if needed
	for(TSelectionFrames::reverse_iterator it(selectionFrames.rbegin()); it != selectionFrames.rend(); it++)
	{
		if ((int)*it < (int)currMovieData.records.size()) break;
		delete_index = *it;
		// reduce selection manually, because reduced list won't call ItemChanged for these rows
		selectionFrames.erase(delete_index);
		if (!selectionFrames.size()) break;
	}
	// reduce greenzone
	InvalidateGreenZone(index);
}

static void ClearFrames()
{
	//clear input on each selection
	for(TSelectionFrames::iterator it(selectionFrames.begin()); it != selectionFrames.end(); it++)
	{
		currMovieData.records[*it].clear();
	}
	InputChanged();
	// reduce greenzone
	int index = *selectionFrames.begin();
	InvalidateGreenZone(index);
}

//the column set operation, for setting a button/Marker for a span of selected values
static void ColumnSet(int column)
{
	if (column == COLUMN_FRAMENUM || column == COLUMN_FRAMENUM2)
	{
		// Markers column
		//inspect the selected frames, if they are all set, then unset all, else set all
		bool unset_found = false;
		for(TSelectionFrames::iterator it(selectionFrames.begin()); it != selectionFrames.end(); it++)
		{
			if(!(currMovieData.frames_flags[*it] & MARKER_FLAG_BIT))
			{
				unset_found = true;
				break;
			}
		}
		if (unset_found)
		{
			// set all
			for(TSelectionFrames::iterator it(selectionFrames.begin()); it != selectionFrames.end(); it++)
				currMovieData.frames_flags[*it] |= MARKER_FLAG_BIT;
		} else
		{
			// unset all
			for(TSelectionFrames::iterator it(selectionFrames.begin()); it != selectionFrames.end(); it++)
				currMovieData.frames_flags[*it] &= ~MARKER_FLAG_BIT;
		}
		MarkersChanged();
		ClearSelection();
		RedrawList();
	} else
	{
		// buttons column
		int joy = (column - COLUMN_JOYPAD1_A) / NUM_JOYPAD_BUTTONS;
		if (joy < 0 || joy >= NUM_JOYPADS) return;
		int button = (column - COLUMN_JOYPAD1_A) % NUM_JOYPAD_BUTTONS;
		//inspect the selected frames, if they are all set, then unset all, else set all
		bool newValue = false;
		for(TSelectionFrames::iterator it(selectionFrames.begin()); it != selectionFrames.end(); it++)
		{
			if(!(currMovieData.records[*it].checkBit(joy,button)))
			{
				newValue = true;
				break;
			}
		}
		// apply newValue
		for(TSelectionFrames::iterator it(selectionFrames.begin()); it != selectionFrames.end(); it++)
			currMovieData.records[*it].setBitValue(joy,button,newValue);
		InputChanged();
		InvalidateGreenZone(*selectionFrames.begin());
	}
}

void ClearSelection()
{
	ListView_SetItemState(hwndList,-1,0, LVIS_SELECTED);
}
void ClearRowSelection(int index)
{
	ListView_SetItemState(hwndList,index,0, LVIS_SELECTED);
}

void SelectAll()
{
	ListView_SetItemState(hwndList,-1,LVIS_SELECTED, LVIS_SELECTED);
	RedrawList();
}
void SelectMidMarkers()
{
	int center, upper_border, lower_border;
	int upper_marker, lower_marker;
	int movie_size = currMovieData.records.size();

	// if selection size=0 then playback cursor is selected and serves as center
	if (selectionFrames.size())
	{
		upper_border = center = *selectionFrames.begin();
		lower_border = *selectionFrames.rbegin();
	} else lower_border = upper_border = center = currFrameCounter;

	// find markers
	// searching up starting from center-0
	for (upper_marker = center; upper_marker >= 0; upper_marker--)
		if (currMovieData.frames_flags[upper_marker] & MARKER_FLAG_BIT) break;
	// searching down starting from center+1
	for (lower_marker = center+1; lower_marker < movie_size; ++lower_marker)
		if (currMovieData.frames_flags[lower_marker] & MARKER_FLAG_BIT) break;
	
	if (upper_marker == -1 && lower_marker == movie_size)
	{
		SelectAll();
		return;
	}

	ClearSelection();
	// selecting circle:
	if (upper_border > upper_marker+1 || lower_border < lower_marker-1 || lower_border > lower_marker)
	{
		// default: select all between markers
		for (int i = upper_marker+1; i < lower_marker; ++i)
		{
			ListView_SetItemState(hwndList,i,LVIS_SELECTED,LVIS_SELECTED);
		}
	} else if (upper_border == upper_marker+1 && lower_border == lower_marker-1)
	{
		// already selected all between markers - now select both markers or at least select the marker that is not outside movie range
		if (upper_marker < 0) upper_marker = 0;
		if (lower_marker >= movie_size) lower_marker >= movie_size - 1;
		for (int i = upper_marker; i <= lower_marker; ++i)
		{
			ListView_SetItemState(hwndList,i,LVIS_SELECTED,LVIS_SELECTED);
		}
	} else if (upper_border <= upper_marker && lower_border >= lower_marker)
	{
		// selected all between markers and both markers selected too - now deselect lower marker
		ListView_SetItemState(hwndList,lower_marker,0,LVIS_SELECTED);
		for (int i = upper_marker; i < lower_marker; ++i)
		{
			ListView_SetItemState(hwndList,i,LVIS_SELECTED,LVIS_SELECTED);
		}
	} else if (upper_border == upper_marker && lower_border == lower_marker-1)
	{
		// selected all between markers and upper marker selected too - now deselect upper marker and (if lower marker < movie_size) reselect lower marker
		ListView_SetItemState(hwndList,upper_marker,0,LVIS_SELECTED);
		if (lower_marker >= movie_size) lower_marker >= movie_size - 1;
		for (int i = upper_marker+1; i <= lower_marker; ++i)
		{
			ListView_SetItemState(hwndList,i,LVIS_SELECTED,LVIS_SELECTED);
		}
	} else if (upper_border == upper_marker+1 && lower_border == lower_marker)
	{
		// selected all between markers and lower marker selected too - now deselect lower marker (return to "selected all between markers")
		ListView_SetItemState(hwndList,lower_marker,0,LVIS_SELECTED);
		for (int i = upper_marker + 1; i < lower_marker; ++i)
		{
			ListView_SetItemState(hwndList,i,LVIS_SELECTED,LVIS_SELECTED);
		}
	}
}

//copies the current selection to the clipboard
static bool Copy()
{
	if (selectionFrames.size()==0) return false;

	int cframe=*selectionFrames.begin()-1;
    try 
	{
		int range = *selectionFrames.rbegin() - *selectionFrames.begin()+1;
		//std::string outbuf clipString("TAS");

		std::stringstream clipString;
		clipString << "TAS " << range << std::endl;

		for(TSelectionFrames::iterator it(selectionFrames.begin()); it != selectionFrames.end(); it++)
		{
			if (*it>cframe+1)
			{
				clipString << '+' << (*it-cframe) << '|';
			}
			cframe=*it;

			int cjoy=0;
			for (int joy=0; joy<NUM_JOYPADS; ++joy)
			{
				while (currMovieData.records[*it].joysticks[joy] && cjoy<joy) 
				{
					clipString << '|';
					++cjoy;
				}
				for (int bit=0; bit<NUM_JOYPAD_BUTTONS; ++bit)
				{
					if (currMovieData.records[*it].joysticks[joy] & (1<<bit))
					{
						clipString << MovieRecord::mnemonics[bit];
					}
				}
			}
			clipString << std::endl;

			if (!OpenClipboard(hwndTasEdit))
				return false;
			EmptyClipboard();

			HGLOBAL hGlobal = GlobalAlloc(GMEM_MOVEABLE, clipString.str().size()+1);

			if (hGlobal==INVALID_HANDLE_VALUE)
			{
				CloseClipboard();
				return false;
			}
			char *pGlobal = (char*)GlobalLock(hGlobal);
			strcpy(pGlobal, clipString.str().c_str());
			GlobalUnlock(hGlobal);
			SetClipboardData(CF_TEXT, hGlobal);

			CloseClipboard();
		}
		
	}
	catch (std::bad_alloc e)
	{
		return false;
	}

	return true;
}

//cuts the current selection, copying it to the clipboard
static void Cut()
{
	if (Copy())
	{
		ClearFrames();
	}
}

//pastes the current clipboard selection into current inputlog
static bool Paste()
{
	bool result = false;
	if (selectionFrames.size()==0)
		return false;

	int pos = *selectionFrames.begin();

	if (!OpenClipboard(hwndTasEdit))
		return false;
	
	HANDLE hGlobal = GetClipboardData(CF_TEXT);
	if (hGlobal)
	{
		char *pGlobal = (char*)GlobalLock((HGLOBAL)hGlobal);

		// TAS recording info starts with "TAS ".
		if (pGlobal[0]=='T' && pGlobal[1]=='A' && pGlobal[2]=='S')
		{
			int range;

			// Extract number of frames
			sscanf (pGlobal+3, "%d", &range);
			if (currMovieData.records.size()<static_cast<unsigned int>(pos+range))
			{
				currMovieData.insertEmpty(currMovieData.records.size(),pos+range-currMovieData.records.size());
			}

			pGlobal = strchr(pGlobal, '\n');
			int joy=0;
			--pos;
			
			while (pGlobal++ && *pGlobal!='\0')
			{
				char *frame = pGlobal;

				// Detect skipped frames in paste.
				if (frame[0]=='+')
				{
					pos += atoi(frame+1);
					while (*frame && *frame != '\n' && *frame!='|')
						++frame;
					if (*frame=='|') ++frame;
				} else
				{
					++pos;
				}

				currMovieData.records[pos].joysticks[0]=0;
				currMovieData.records[pos].joysticks[1]=0;
				int joy=0;

				while (*frame && *frame != '\n' && *frame !='\r')
				{
					switch (*frame)
					{
					case '|': // Joystick marker
						++joy;
						break;
					default:
						for (int bit=0; bit<NUM_JOYPAD_BUTTONS; ++bit)
						{
							if (*frame==MovieRecord::mnemonics[bit])
							{
								currMovieData.records[pos].joysticks[joy]|=(1<<bit);
								break;
							}
						}
						break;
					}
					++frame;
				}

				pGlobal = strchr(pGlobal, '\n');
			}
			InputChanged();
			InvalidateGreenZone(*selectionFrames.begin());
			result = true;
		}

		GlobalUnlock(hGlobal);
	}

	CloseClipboard();
	return result;
}

// ---------------------------------------------------------------------------------
//The subclass wndproc for the listview header
static LRESULT APIENTRY HeaderWndProc(HWND hWnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
	switch(msg)
	{
	case WM_SETCURSOR:
		return true;	// no column resizing
	case WM_LBUTTONDOWN:
	case WM_LBUTTONDBLCLK:
		{
			if (selectionFrames.size())
			{
				//perform hit test
				HD_HITTESTINFO info;
				info.pt.x = GET_X_LPARAM(lParam);
				info.pt.y = GET_Y_LPARAM(lParam);
				SendMessage(hWnd,HDM_HITTEST,0,(LPARAM)&info);
				if(info.iItem >= COLUMN_FRAMENUM && info.iItem <= COLUMN_FRAMENUM2)
					ColumnSet(info.iItem);
			}
		}
		return true;
	}
	return CallWindowProc(hwndHeader_oldWndproc,hWnd,msg,wParam,lParam);
}

//The subclass wndproc for the listview
static LRESULT APIENTRY ListWndProc(HWND hWnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
	switch(msg)
	{
	case WM_CHAR:
		return 0;
	case WM_KILLFOCUS:
		return 0;
	case WM_NOTIFY:
		{
			switch (((LPNMHDR)lParam)->code)
			{
			case HDN_BEGINTRACKW:
			case HDN_BEGINTRACKA:
			case HDN_TRACK:
				return true;	// no column resizing
			}
			break;
		}

	}
	return CallWindowProc(hwndList_oldWndProc,hWnd,msg,wParam,lParam);
}

//All dialog initialization
static void InitDialog()
{
	//prepare the listview
	ListView_SetExtendedListViewStyleEx(hwndList,
                             LVS_EX_FULLROWSELECT | LVS_EX_GRIDLINES ,
                             LVS_EX_FULLROWSELECT | LVS_EX_GRIDLINES );

	//subclass the header
	hwndHeader = ListView_GetHeader(hwndList);
	hwndHeader_oldWndproc = (WNDPROC)SetWindowLong(hwndHeader,GWL_WNDPROC,(LONG)HeaderWndProc);

	//subclass the whole listview, so we can block some keystrokes
	hwndList_oldWndProc = (WNDPROC)SetWindowLong(hwndList,GWL_WNDPROC,(LONG)ListWndProc);

	//setup all images for the listview
	HIMAGELIST himglist = ImageList_Create(8,12,ILC_COLOR8 | ILC_MASK,1,1);
	HBITMAP bmp = LoadBitmap(fceu_hInstance,MAKEINTRESOURCE(IDB_TE_ARROW));
	ImageList_AddMasked(himglist, bmp, 0xFF00FF);
	DeleteObject(bmp);
	ListView_SetImageList(hwndList,himglist,LVSIL_SMALL);
	//doesnt work well??
	//HIMAGELIST himglist = ImageList_LoadImage(fceu_hInstance,MAKEINTRESOURCE(IDB_TE_ARROW),12,1,RGB(255,0,255),IMAGE_BITMAP,LR_DEFAULTCOLOR);

	//setup columns
	LVCOLUMN lvc;
	int colidx=0;
	// icons column
	lvc.mask = LVCF_WIDTH;
	lvc.cx = 12;
	ListView_InsertColumn(hwndList, colidx++, &lvc);
	// frame number column
	lvc.mask = LVCF_WIDTH | LVCF_TEXT | LVCF_FMT;
	lvc.fmt = LVCFMT_CENTER;
	lvc.cx = 75;
	lvc.pszText = "Frame#";
	ListView_InsertColumn(hwndList, colidx++, &lvc);
	// pads columns
	lvc.cx = 21;
	// add pads 1 and 2
	for (int joy = 0; joy < 2; ++joy)
	{
		for (int btn = 0; btn < NUM_JOYPAD_BUTTONS; ++btn)
		{
			lvc.pszText = buttonNames[btn];
			ListView_InsertColumn(hwndList, colidx++, &lvc);
		}
	}
	// add pads 3 and 4 and frame_number2
	if (currMovieData.fourscore) AddFourscore();

	//the initial update
	UpdateTasEdit();
}

void AddFourscore()
{
	// add list columns
	LVCOLUMN lvc;
	lvc.mask = LVCF_WIDTH | LVCF_TEXT | LVCF_FMT;
	lvc.fmt = LVCFMT_CENTER;
	lvc.cx = 21;
	int colidx = COLUMN_JOYPAD3_A;
	for (int joy = 0; joy < 2; ++joy)
	{
		for (int btn = 0; btn < NUM_JOYPAD_BUTTONS; ++btn)
		{
			lvc.pszText = buttonNames[btn];
			ListView_InsertColumn(hwndList, colidx++, &lvc);
		}
	}
	// frame number column again
	lvc.cx = 75;
	lvc.pszText = "Frame#";
	ListView_InsertColumn(hwndList, colidx++, &lvc);
	// enable radiobuttons for 3P/4P multitracking
	EnableWindow(hwndRB_Rec3P, true);
	EnableWindow(hwndRB_Rec4P, true);
	// change eoptions
	FCEUI_SetInputFourscore(true);


}
void RemoveFourscore()
{
	// remove list columns
	for (int i = COLUMN_FRAMENUM2; i >= COLUMN_JOYPAD3_A; --i)
	{
		ListView_DeleteColumn (hwndList, i);
	}
	// disable radiobuttons for 3P/4P multitracking
	EnableWindow(hwndRB_Rec3P, false);
	EnableWindow(hwndRB_Rec4P, false);
	// change eoptions
	FCEUI_SetInputFourscore(false);


}

void UncheckRecordingRadioButtons()
{
	Button_SetCheck(hwndRB_RecOff, BST_UNCHECKED);
	Button_SetCheck(hwndRB_RecAll, BST_UNCHECKED);
	Button_SetCheck(hwndRB_Rec1P, BST_UNCHECKED);
	Button_SetCheck(hwndRB_Rec2P, BST_UNCHECKED);
	Button_SetCheck(hwndRB_Rec3P, BST_UNCHECKED);
	Button_SetCheck(hwndRB_Rec4P, BST_UNCHECKED);
}
void RecheckRecordingRadioButtons()
{
	old_movie_readonly = movie_readonly;
	old_multitrack_recording_joypad = multitrack_recording_joypad;
	if (movie_readonly)
	{
		Button_SetCheck(hwndRB_RecOff, BST_CHECKED);
	} else
	{
		switch(multitrack_recording_joypad)
		{
		case MULTITRACK_RECORDING_ALL:
			Button_SetCheck(hwndRB_RecAll, BST_CHECKED);
			break;
		case MULTITRACK_RECORDING_1P:
			Button_SetCheck(hwndRB_Rec1P, BST_CHECKED);
			break;
		case MULTITRACK_RECORDING_2P:
			Button_SetCheck(hwndRB_Rec2P, BST_CHECKED);
			break;
		case MULTITRACK_RECORDING_3P:
			Button_SetCheck(hwndRB_Rec3P, BST_CHECKED);
			break;
		case MULTITRACK_RECORDING_4P:
			Button_SetCheck(hwndRB_Rec4P, BST_CHECKED);
			break;
		}
	}
}

//Creates a new TASEdit Project
static void NewProject()
{
	//determine if current project changed
	//if so, ask to save changes
	//close current project
	if (!AskSaveProject()) return;

	//TODO: close current project instance, create a new one with a non-parameterized constructor
}

//Opens a new Project file
void OpenProject()
{
	if (!AskSaveProject()) return;
		
	//If OPENFILENAME dialog successful, open up a completely new project instance and scrap the old one
	//Run the project Load() function to pull all info from the .tas file into this new project instance

	const char TPfilter[]="TASEdit Project (*.tas)\0*.tas\0\0";	

	OPENFILENAME ofn;								
	memset(&ofn,0,sizeof(ofn));						
	ofn.lStructSize=sizeof(ofn);					
	ofn.hInstance=fceu_hInstance;
	ofn.lpstrTitle="Open TASEdit Project...";
	ofn.lpstrFilter=TPfilter;

	char nameo[2048];								//File name
	strcpy(nameo, GetRomName());					//For now, just use ROM name

	ofn.lpstrFile=nameo;							
	ofn.nMaxFile=256;
	ofn.Flags=OFN_EXPLORER|OFN_HIDEREADONLY|OFN_OVERWRITEPROMPT|OFN_FILEMUSTEXIST;
	string initdir =  FCEU_GetPath(FCEUMKF_MOVIE);	
	ofn.lpstrInitialDir=initdir.c_str();

	if(GetOpenFileName(&ofn))							//If it is a valid filename
	{							
		std::string tempstr = nameo;					//Make a temporary string for filename
		char drv[512], dir[512], name[512], ext[512];	//For getting the filename!
		if(tempstr.rfind(".tas") == std::string::npos)	//If they haven't put ".tas" after it, stick it on ourselves
			tempstr.append(".tas");						
		splitpath(tempstr.c_str(), drv, dir, name, ext);	//Split the path...
		project.SetProjectName(name);
		std::string filename = name;						//Get the filename
		filename.append(ext);								//Shove the extension back onto it...
		project.SetProjectFile(filename);					//And update the project's filename.
		//Set the fm2 name
		std::string thisfm2name = name;
		thisfm2name.append(".fm2");
		project.SetFM2Name(thisfm2name);
		// switch to read-only mode, but first must uncheck radiobuttons explicitly
		if (!movie_readonly) FCEUI_MovieToggleReadOnly();
		multitrack_recording_joypad = MULTITRACK_RECORDING_ALL;
		UncheckRecordingRadioButtons();
		RecheckRecordingRadioButtons();
		// remember to update fourscore status
		bool last_fourscore = currMovieData.fourscore;
		// Load project
		project.LoadProject(project.GetProjectFile());
		if (last_fourscore && !currMovieData.fourscore)
			RemoveFourscore();
		else if (!last_fourscore && currMovieData.fourscore)
			AddFourscore();
		SeekingStop();
		FollowPlayback();
		//UpdateTasEdit();
		RedrawTasedit();
		RedrawWindowCaption();
	}
}

// Saves current project
bool SaveProjectAs()
{
//Save project as new user selected filename
//flag project as not changed

	const char TPfilter[]="TASEdit Project (*.tas)\0*.tas\0All Files (*.*)\0*.*\0\0";	//Filetype filter

	OPENFILENAME ofn;								
	memset(&ofn,0,sizeof(ofn));						
	ofn.lStructSize=sizeof(ofn);					
	ofn.hInstance=fceu_hInstance;
	ofn.lpstrTitle="Save TASEdit Project As...";
	ofn.lpstrFilter=TPfilter;

	char nameo[2048];										//File name
	strcpy(nameo, GetRomName());							//For now, just use ROM name

	ofn.lpstrFile=nameo;									//More parameters
	ofn.lpstrDefExt="tas";
	ofn.nMaxFile=256;
	ofn.Flags=OFN_EXPLORER|OFN_HIDEREADONLY|OFN_OVERWRITEPROMPT;
	string initdir =  FCEU_GetPath(FCEUMKF_MOVIE);			//Initial directory
	ofn.lpstrInitialDir=initdir.c_str();

	if(GetSaveFileName(&ofn))								//If it is a valid filename
	{
		std::string tempstr = nameo;						//Make a temporary string for filename
		char drv[512], dir[512], name[512], ext[512];		//For getting the filename!

		splitpath(tempstr.c_str(), drv, dir, name, ext);	//Split it up...
		project.SetProjectName(name);
		std::string filename = name;						//Grab the name...
		filename.append(ext);								//Stick extension back on...
		project.SetProjectFile(filename);					//And update the project's filename.
		std::string thisfm2name = name;
		thisfm2name.append(".fm2");							//Setup the fm2 name
		project.SetFM2Name(thisfm2name);					//Set the project's fm2 name
		project.saveProject();
	} else return false;
	// saved successfully - remove * mark from caption
	RedrawWindowCaption();
	return true;
}
bool SaveProject()
{
	if (!project.saveProject())
		return SaveProjectAs();
	else
		RedrawWindowCaption();
	return true;
}

// returns false if user doesn't want to exit
bool AskSaveProject()
{
	bool changes_found = false;
	if (project.changed) changes_found = true;

	// ask saving project
	if (changes_found)
	{
		int answer = MessageBox(hwndTasEdit, "Save Project changes?", "TAS Edit", MB_YESNOCANCEL);
		if(answer == IDYES)
			return SaveProject();
		return (answer != IDCANCEL);
	}
	return true;
}

//Takes a selected .fm2 file and adds it to the Project inputlog
static void Import()
{
	//Pull the fm2 header, comments, subtitle information out and put it into the project info
	//Pull the input out and add it to the main branch input log file
}

//Takes current inputlog and saves it as a .fm2 file
static void Export()
{
	//TODO: redesign this
	//Dump project header info into file, then comments & subtitles, then input log
	//This will require special prunctions, ::DumpHeader  ::DumpComments etc
	const char filter[]="FCEUX Movie File (*.fm2)\0*.fm2\0All Files (*.*)\0*.*\0\0";
	char fname[2048] = {0};
	OPENFILENAME ofn;
	memset(&ofn,0,sizeof(ofn));
	ofn.lStructSize=sizeof(ofn);
	ofn.hInstance=fceu_hInstance;
	ofn.lpstrTitle="Export TAS as...";
	ofn.lpstrFilter=filter;
	ofn.lpstrFile=fname;
	ofn.lpstrDefExt="fm2";
	ofn.nMaxFile=256;
	std::string initdir = FCEU_GetPath(FCEUMKF_MOVIE);
	ofn.lpstrInitialDir=initdir.c_str();
	if(GetSaveFileName(&ofn))
	{
		EMUFILE* osRecordingMovie = FCEUD_UTF8_fstream(fname, "wb");
		currMovieData.dump(osRecordingMovie,false);
		delete osRecordingMovie;
		osRecordingMovie = 0;
	}
}

static void Truncate()
{
	int frame = currFrameCounter;
	if (selectionFrames.size())
	{
		frame=*selectionFrames.begin();
		JumpToFrame(frame);
		ClearSelection();
	}
	currMovieData.truncateAt(frame+1);
	InputChanged();
	UpdateList();
	InvalidateGreenZone(frame);
}

//likewise, handles a changed item range from the listview
static void ItemRangeChanged(NMLVODSTATECHANGE* info)
{
	bool ON = !(info->uOldState & LVIS_SELECTED) && (info->uNewState & LVIS_SELECTED);
	bool OFF = (info->uOldState & LVIS_SELECTED) && !(info->uNewState & LVIS_SELECTED);

	if(ON)
		for(int i=info->iFrom;i<=info->iTo;i++)
			selectionFrames.insert(i);
	else
		for(int i=info->iFrom;i<=info->iTo;i++)
			selectionFrames.erase(i);
}

//handles a changed item from the listview
//used to track selection
static void ItemChanged(NMLISTVIEW* info)
{
	int item = info->iItem;
	
	bool ON = !(info->uOldState & LVIS_SELECTED) && (info->uNewState & LVIS_SELECTED);
	bool OFF = (info->uOldState & LVIS_SELECTED) && !(info->uNewState & LVIS_SELECTED);

	//if the item is -1, apply the change to all items
	if(item == -1)
	{
		if(OFF)
		{
			// clear all
			selectionFrames.clear();
		} else if (ON)
		{
			// select all
			for(int i = 0; i < currMovieData.records.size(); i++)
			{
				selectionFrames.insert(i);
			}
		}
	}
	else
	{
		if(ON)
			selectionFrames.insert(item);
		else if(OFF) 
			selectionFrames.erase(item);
	}
}

BOOL CALLBACK WndprocTasEdit(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch(uMsg)
	{
		case WM_PAINT:
			{
			}
			break;
		case WM_INITDIALOG:
			if (TasEdit_wndx==-32000) TasEdit_wndx=0; //Just in case
			if (TasEdit_wndy==-32000) TasEdit_wndy=0;
			SetWindowPos(hwndDlg,0,TasEdit_wndx,TasEdit_wndy,0,0,SWP_NOSIZE|SWP_NOZORDER|SWP_NOOWNERZORDER);
			// save references to dialog items
			hwndList = GetDlgItem(hwndDlg, IDC_LIST1);
			listItems = ListView_GetCountPerPage(hwndList);
			FCEU_printf("listItems = %d\n\n",listItems);

			hwndProgressbar = GetDlgItem(hwndDlg, IDC_PROGRESS1);
			SendMessage(hwndProgressbar, PBM_SETRANGE, 0, MAKELPARAM(0, PROGRESSBAR_WIDTH)); 
			hwndRewind = GetDlgItem(hwndDlg, TASEDIT_REWIND);
			hwndForward = GetDlgItem(hwndDlg, TASEDIT_FORWARD);
			hwndRB_RecOff = GetDlgItem(hwndDlg, IDC_RADIO1);
			hwndRB_RecAll = GetDlgItem(hwndDlg, IDC_RADIO2);
			hwndRB_Rec1P = GetDlgItem(hwndDlg, IDC_RADIO3);
			hwndRB_Rec2P = GetDlgItem(hwndDlg, IDC_RADIO4);
			hwndRB_Rec3P = GetDlgItem(hwndDlg, IDC_RADIO5);
			hwndRB_Rec4P = GetDlgItem(hwndDlg, IDC_RADIO6);
			InitDialog();
			break; 

		case WM_MOVE:
			{
				if (!IsIconic(hwndDlg))
				{
					RECT wrect;
					GetWindowRect(hwndDlg,&wrect);
					TasEdit_wndx = wrect.left;
					TasEdit_wndy = wrect.top;

					WindowBoundsCheckNoResize(TasEdit_wndx,TasEdit_wndy,wrect.right);
				}
				break;
			}

		case WM_NOTIFY:

			switch(wParam)
			{
			case IDC_LIST1:
				switch(((LPNMHDR)lParam)->code)
				{
				case NM_CUSTOMDRAW:
					SetWindowLong(hwndDlg, DWL_MSGRESULT, CustomDraw((NMLVCUSTOMDRAW*)lParam));
					return TRUE;
				case LVN_GETDISPINFO:
					GetDispInfo((NMLVDISPINFO*)lParam);
					break;
				case NM_CLICK:
					SingleClick((LPNMITEMACTIVATE)lParam);
					break;
				case NM_DBLCLK:
					DoubleClick((LPNMITEMACTIVATE)lParam);
					break;
				case NM_RCLICK:
					RightClick((LPNMITEMACTIVATE)lParam);
					break;
				case LVN_ITEMCHANGED:
					ItemChanged((LPNMLISTVIEW) lParam);
					break;
				case LVN_ODSTATECHANGED:
					ItemRangeChanged((LPNMLVODSTATECHANGE) lParam);
					break;
					/*
				case LVN_ENDSCROLL:
					// redraw upper and lower list rows (fix for known WinXP bug)
					int start = ListView_GetTopIndex(hwndList);
					ListView_RedrawItems(hwndList,start,start);
					int end = start + listItems - 1;
					ListView_RedrawItems(hwndList,end,end);
					break;
					*/

				}
				break;
			case TASEDIT_PLAYSTOP:
				switch(((LPNMHDR)lParam)->code)
				{
				case NM_CLICK:
				case NM_DBLCLK:
					Tasedit_ToggleEmulationPause();
					break;
				}
				break;
			}
			break;
		
		case WM_CLOSE:
		case WM_QUIT:
			ExitTasEdit();
			break;

		case WM_ACTIVATEAPP:
			if((BOOL)wParam)
				TASEdit_focus = true;
			else
				TASEdit_focus = false;
			return DefWindowProc(hwndDlg,uMsg,wParam,lParam);

		case WM_COMMAND:
			switch(LOWORD(wParam))
			{
			case ID_FILE_NEWPROJECT:
				NewProject();
				break;
			case ID_FILE_OPENPROJECT:
				OpenProject();
				break;
			case ACCEL_CTRL_S:
			case ID_FILE_SAVEPROJECT:
					SaveProject();
				break;
			case ID_FILE_SAVEPROJECTAS:
					SaveProjectAs();
				break;
			case ID_FILE_IMPORTFM2:
				Replay_LoadMovie(true);
				//Import(); //adelikat:  Putting the play movie dialog in its place until the import concept is refined.
				break;
			case ID_FILE_EXPORTFM2:
					Export();
				break;
			case ID_TASEDIT_FILE_CLOSE:
				ExitTasEdit();
				break;
			case ID_EDIT_SELECTALL:
				SelectAll();
				break;
			case ACCEL_CTRL_X:
			case ID_TASEDIT_CUT:
				Cut();
				break;
			case ACCEL_CTRL_C:
			case ID_TASEDIT_COPY:
				Copy();
				break;
			case ACCEL_CTRL_V:
			case ID_TASEDIT_PASTE:
				Paste();
				break;
			case ACCEL_CTRL_DELETE:
			case ID_TASEDIT_DELETE:
			case ID_CONTEXT_SELECTED_DELETEFRAMES:
				if (selectionFrames.size()) DeleteFrames();
				break;
			case ACCEL_CTRL_T:
			case ID_EDIT_TRUNCATE:
			case ID_CONTEXT_SELECTED_TRUNCATE:
			case ID_CONTEXT_STRAY_TRUNCATE:
				Truncate();
				break;
			case ID_HELP_TASEDITHELP:
				OpenHelpWindow(tasedithelp);
				//link to TASEdit in help menu
				break;
			case ACCEL_INS:
			case ID_EDIT_INSERT:
			case MENU_CONTEXT_STRAY_INSERTFRAMES:
			case ID_CONTEXT_SELECTED_INSERTFRAMES2:
				{
					int frames = selectionFrames.size();
					if(CWin32InputBox::GetInteger("Insert number of Frames", "How many frames?", frames, hwndDlg) == IDOK)
					{
						if (frames > 0)
						{
							if (selectionFrames.size())
							{
								// insert at selection
								int index = *selectionFrames.begin();
								currMovieData.insertEmpty(index,frames);
								InputChanged();
								InvalidateGreenZone(index);
							} else
							{
								// insert at playback cursor
								currMovieData.insertEmpty(currFrameCounter,frames);
								InputChanged();
								InvalidateGreenZone(currFrameCounter);
							}
						}
					}
				}
				break;
			case ACCEL_CTRL_INSERT:
			case ID_EDIT_INSERTFRAMES:
			case ID_CONTEXT_SELECTED_INSERTFRAMES:
				if (selectionFrames.size()) InsertFrames();
				break;
			case ACCEL_DEL:
			case ID_EDIT_CLEAR:
			case ID_CONTEXT_SELECTED_CLEARFRAMES:
				if (selectionFrames.size()) ClearFrames();
				break;
			case TASEDIT_REWIND_FULL:
				//rewinds to beginning of greenzone
				JumpToFrame(FindBeginningOfGreenZone(0));
				FollowPlayback();
				break;
			case TASEDIT_PLAYSTOP:
				Tasedit_ToggleEmulationPause();
				break;
			case TASEDIT_FORWARD_FULL:
				//moves to the end of greenzone
				JumpToFrame(currMovieData.greenZoneCount-1);
				FollowPlayback();
				break;
			case ACCEL_CTRL_F:
			case CHECK_FOLLOW_CURSOR:
				//switch "Follow playback" flag
				TASEdit_follow_playback ^= 1;
				CheckDlgButton(hwndTasEdit, CHECK_FOLLOW_CURSOR, TASEdit_follow_playback?MF_CHECKED : MF_UNCHECKED);
				// if switched off then jump to selection
				if (TASEdit_follow_playback)
					FollowPlayback();
				else if (selectionFrames.size())
					ListView_EnsureVisible(hwndList,(int)*selectionFrames.begin(),FALSE);
				break;
			case ID_VIEW_SHOW_LAG_FRAMES:
				//switch "Highlight lag frames" flag
				TASEdit_show_lag_frames ^= 1;
				CheckMenuItem(hmenu, ID_VIEW_SHOW_LAG_FRAMES, TASEdit_show_lag_frames?MF_CHECKED : MF_UNCHECKED);
				RedrawList();
				break;
			case ID_VIEW_SHOW_MARKERS:
				//switch "Show Markers" flag
				TASEdit_show_markers ^= 1;
				CheckMenuItem(hmenu, ID_VIEW_SHOW_MARKERS, TASEdit_show_markers?MF_CHECKED : MF_UNCHECKED);
				RedrawList();
				break;
			case ID_VIEW_SHOWDOTINEMPTYCELLS:
				TASEdit_show_dot ^= 1;
				CheckMenuItem(hmenu, ID_VIEW_SHOWDOTINEMPTYCELLS, TASEdit_show_dot?MF_CHECKED : MF_UNCHECKED);
				RedrawList();
				break;
			case ACCEL_CTRL_P:
			case CHECK_AUTORESTORE_PLAYBACK:
				//switch "Auto-restore last playback position" flag
				TASEdit_restore_position ^= 1;
				CheckDlgButton(hwndTasEdit,CHECK_AUTORESTORE_PLAYBACK,TASEdit_restore_position?BST_CHECKED:BST_UNCHECKED);
				break;
			case ID_CONFIG_SETGREENZONECAPACITY:
				{
					//open input dialog
					int new_capacity = TASEdit_greenzone_capacity;
					if(CWin32InputBox::GetInteger("Greenzone capacity", "Keep savestates for how many frames?", new_capacity,  hwndDlg) == IDOK)
					{
						if (new_capacity < GREENZONE_MIN_CAPACITY)
							new_capacity = GREENZONE_MIN_CAPACITY;
						else if (new_capacity > GREENZONE_MAX_CAPACITY)
							new_capacity = GREENZONE_MAX_CAPACITY;
						if (new_capacity < TASEdit_greenzone_capacity)
						{
							TASEdit_greenzone_capacity = new_capacity;
							currMovieData.ClearGreenzoneTail();
							RedrawList();
						} else TASEdit_greenzone_capacity = new_capacity;
					}
					break;
				}
			case ID_CONFIG_MUTETURBO:
				muteTurbo ^= 1;
				CheckMenuItem(hmenu, ID_CONFIG_MUTETURBO, muteTurbo?MF_CHECKED : MF_UNCHECKED);
				break;
			case ID_CONFIG_BINDMARKERSTOINPUT:
				//switch "Bind Markers to Input" flag
				TASEdit_bind_markers ^= 1;
				CheckMenuItem(hmenu, ID_CONFIG_BINDMARKERSTOINPUT, TASEdit_bind_markers?MF_CHECKED : MF_UNCHECKED);
				break;
			case IDC_PROGRESS_BUTTON:
				if (pauseframe) SeekingStop();
				break;
			case IDC_RADIO1:
				// switch to readonly, no need to recheck radiobuttons
				if (!movie_readonly) FCEUI_MovieToggleReadOnly();
				break;
			case IDC_RADIO2:
				// switch to read+write for all, no need to recheck radiobuttons
				if (movie_readonly) FCEUI_MovieToggleReadOnly();
				multitrack_recording_joypad = MULTITRACK_RECORDING_ALL;
				break;
			case IDC_RADIO3:
				// switch to read+write for 1P, no need to recheck radiobuttons
				if (movie_readonly) FCEUI_MovieToggleReadOnly();
				multitrack_recording_joypad = MULTITRACK_RECORDING_1P;
				break;
			case IDC_RADIO4:
				// switch to read+write for 2P, no need to recheck radiobuttons
				if (movie_readonly) FCEUI_MovieToggleReadOnly();
				multitrack_recording_joypad = MULTITRACK_RECORDING_2P;
				break;
			case IDC_RADIO5:
				// switch to read+write for 3P, no need to recheck radiobuttons
				if (movie_readonly) FCEUI_MovieToggleReadOnly();
				multitrack_recording_joypad = MULTITRACK_RECORDING_3P;
				break;
			case IDC_RADIO6:
				// switch to read+write for 4P, no need to recheck radiobuttons
				if (movie_readonly) FCEUI_MovieToggleReadOnly();
				multitrack_recording_joypad = MULTITRACK_RECORDING_4P;
				break;
			case ACCEL_CTRL_A:
			case ID_EDIT_SELECTMIDMARKERS:
			case ID_SELECTED_SELECTMIDMARKERS:
				SelectMidMarkers();
				break;
			case ACCEL_SHIFT_INS:
			case ID_EDIT_CLONEFRAMES:
			case ID_SELECTED_CLONE:
				if (selectionFrames.size()) CloneFrames();
				break;

			}
			break;

		default:
			break;
	}
	return FALSE;
}

int FindBeginningOfGreenZone(int starting_index)
{
	for (int i=starting_index; i<currMovieData.greenZoneCount; ++i)
	{
		if (!currMovieData.savestates[i].empty()) return i;
	}
	return starting_index;
}
void FollowPlayback()
{
	if (TASEdit_follow_playback) ListView_EnsureVisible(hwndList,currFrameCounter,FALSE);
}

void EnterTasEdit()
{
	if(!FCEU_IsValidUI(FCEUI_TASEDIT)) return;
	// window stuff
	if(!hwndTasEdit) hwndTasEdit = CreateDialog(fceu_hInstance,"TASEDIT",hAppWnd,WndprocTasEdit);
	if(hwndTasEdit)
	{
		// save "eoptions"
		saved_eoptions = eoptions;
		// clear "Run in background"
		eoptions &= ~EO_BGRUN;
		// set "Background TASEdit input"
		KeyboardSetBackgroundAccessBit(KEYBACKACCESS_TASEDIT);
		JoystickSetBackgroundAccessBit(JOYBACKACCESS_TASEDIT);
		// "Set high-priority thread"
		eoptions |= EO_HIGHPRIO;
		DoPriority();
		// clear "Disable speed throttling"
		eoptions &= ~EO_NOTHROTTLE;
		// switch off autosaves
		saved_EnableAutosave = EnableAutosave;
		EnableAutosave = 0;
		// switch on savestates compression
		saved_compressSavestates = compressSavestates;
		compressSavestates = true;


		UpdateCheckedMenuItems();

		hmenu = GetMenu(hwndTasEdit);
		hrmenu = LoadMenu(fceu_hInstance,"TASEDITCONTEXTMENUS");
		// check option ticks
		CheckDlgButton(hwndTasEdit, CHECK_FOLLOW_CURSOR, TASEdit_follow_playback?MF_CHECKED : MF_UNCHECKED);
		CheckMenuItem(hmenu, ID_VIEW_SHOW_LAG_FRAMES, TASEdit_show_lag_frames?MF_CHECKED : MF_UNCHECKED);
		CheckMenuItem(hmenu, ID_VIEW_SHOW_MARKERS, TASEdit_show_markers?MF_CHECKED : MF_UNCHECKED);
		CheckMenuItem(hmenu, ID_CONFIG_BINDMARKERSTOINPUT, TASEdit_bind_markers?MF_CHECKED : MF_UNCHECKED);
		CheckDlgButton(hwndTasEdit,CHECK_AUTORESTORE_PLAYBACK,TASEdit_restore_position?BST_CHECKED:BST_UNCHECKED);
		CheckMenuItem(hmenu, ID_CONFIG_MUTETURBO, muteTurbo?MF_CHECKED : MF_UNCHECKED);
		CheckMenuItem(hmenu, ID_VIEW_SHOWDOTINEMPTYCELLS, TASEdit_show_dot?MF_CHECKED : MF_UNCHECKED);

		SetWindowPos(hwndTasEdit,HWND_TOP,0,0,0,0,SWP_NOSIZE|SWP_NOMOVE|SWP_NOOWNERZORDER);
	}
	project.init();
	// either start new movie or use current movie
	if (movieMode == MOVIEMODE_INACTIVE)
	{
		FCEUI_StopMovie();
		CreateCleanMovie();
		StartFromZero();
	}
	else
	{
		//use current movie to create a new project
		FCEUI_StopMovie();
	}
	// always start work from read-only mode, ready to switch to MULTITRACK_RECORDING_ALL
	movie_readonly = true;
	multitrack_recording_joypad = MULTITRACK_RECORDING_ALL;
	RecheckRecordingRadioButtons();
	movieMode = MOVIEMODE_TASEDIT;
	// init variables
	lastCursor = -1;
	old_project_changed = false;
	old_pauseframe = 0;
	old_show_pauseframe = show_pauseframe = false;
	old_rewind_button_state = rewind_button_state = false;
	old_forward_button_state = forward_button_state = false;
	old_emu_paused = emu_paused = true;
	SeekingStop();
	currMovieData.TryDumpIncremental();
	FCEU_DispMessage("Tasedit engaged",0);
}
bool ExitTasEdit()
{
	if (!AskSaveProject()) return false;

	DestroyWindow(hwndTasEdit);
	hwndTasEdit = 0;
	SeekingStop();
	TASEdit_focus = false;
	// restore "eoptions"
	eoptions = saved_eoptions;
	// restore autosaves
	EnableAutosave = saved_EnableAutosave;
	// restore compression
	compressSavestates = saved_compressSavestates;

	DoPriority();
	UpdateCheckedMenuItems();
	// clear "Background TASEdit input"
	KeyboardClearBackgroundAccessBit(KEYBACKACCESS_TASEDIT);
	JoystickClearBackgroundAccessBit(JOYBACKACCESS_TASEDIT);
	// release memory
	currMovieData.clearGreenzone();
	movieMode = MOVIEMODE_INACTIVE;
	FCEU_DispMessage("Tasedit disengaged",0);
	CreateCleanMovie();
	return true;
}
